# Hands-On #4: Performance counters and memory bandwidth

## Investigate the benchmark code
*LD*: Load, *ST*: Store, *WA*: Write-allocate

- **init**: 0 LD, 1 ST per iteration (+1 LD with WA)
- **copy**: 1 LD, 1 ST per iteration (+1 LD with WA)
- **update**: 1 LD, 1 ST per iteration
- **triad**: 2 LD, 1 ST per iteration (+1 LD with WA)
- **daxpy**: 2 LD, 1 ST per iteration
- **striad**: 3 LD, 1 ST per iteration (+1 LD with WA)
- **sdaxpy**: 3 LD, 1 ST

## Data traffic analysis
While the output of `bwBench` shows us the application bandwidth, the actual hardware bandwidth is might differ
due to write-allocates.
To get the actual sustained bandwidth, we would need to multiply with the factors:

- **init**: 2
- **copy**: 1.5
- **triad**: 1.33
- **striad**: 1.25

Note that newer Intel architectures are more advanced and can overcome write-allocate themselves by using (intrinsic) non-temporal stores.
This results in bandwidths that do not differ by the exact factor mentioned above, but are much closer to the actual hardware bandwidth.

## likwid-perfctr
From the output of `likwid-perfctr`, we can derive the metrics in the bottom "Metric" section:

|              Metric             |  Init |  Copy | Update | Triad | Daxpy | STriad | SDaxpy |
|---------------------------------|-------|-------|--------|-------|-------|--------|--------|
|**Memory read data volume [GB]** | 1.4417| 3.3650| 3.0661 | 6.5287| 6.2963| 9.8095 | 9.5052 |
|**Memory write data volume [GB]**| 2.9077| 3.4367| 2.8930 | 3.2974| 3.0747| 3.2741 | 3.1170 |
|**Read/Write ratio**             |  1:2  |   1:1 |   1:1  |  2:1  |  2:1  |   3:1  |   3:1  |
|**Memory data volume [GB]**      | 4.3494| 6.8017| 5.9591 | 9.8261| 9.3710|13.0836 |12.6222 |
|**Memory bandwidth [GB/s]**      |71.3263|73.6581|73.6534 |77.6599|73.4511|78.6254 |75.1572 |

We can see, the load to store ratio meets our expectations from section 1.
Furthermore, we can now see that the measured hardware bandwidths match (better) the application bandwidth times the factor from section 2.
If the hardware already uses write-allocate evasion (as for Fritz), only the benchmarks like `init` and `copy` show significant difference to the previously reported application bandwidth.


## ccNUMA bandwidth scalability
Since we initialize the data before executing the kernels in the same parallel fashion as we distribute the work later,
the first-touch-policy takes care that all data processed by a thread is in the same ccNUMA domain.
Therefore, we see almost perfect linear scaling in terms of bandwidth from one to four ccNUMA domains in use.
