# Hands-on #2: The divide instruction

## Can you estimate the inverse throughput for a divide operation in CPU cycles?

Input:
* Number of divides:  <img src="https://render.githubusercontent.com/render/math?math=N_{\mathrm{div}}=2\cdot 10^{9}">
* Runtime <img src="https://render.githubusercontent.com/render/math?math=t=1.674\,s">
* Frequency <img src="https://render.githubusercontent.com/render/math?math=f=2.4\cdot 10^{9}\,\tfrac{\mathrm{cy}}{s}">

Question: <img src="https://render.githubusercontent.com/render/math?math=\dfrac{\mathrm{cy}}{\mathrm{divide}}">

Solution: <img src="https://render.githubusercontent.com/render/math?math=\dfrac{\mathrm{cy}}{N_{\mathrm{div}}}=\dfrac{t \cdot f}{N_{\mathrm{div}}}=\dfrac{1.674\,s \cdot 2.4 \cdot 10^{9}\,\mathrm{cy}}{2 \cdot 10^{9}\,s}\approx 2\,\dfrac{\mathrm{cy}}{\mathrm{div}}">


## Now compile successively with different options.

### How does the divide throughput change?

* `-Ofast -xSSE4.2`  
   Runtime `t = 1.748s`  
   Solution: <img src="https://render.githubusercontent.com/render/math?math=\dfrac{\mathrm{cy}}{N_{\mathrm{div}}}=\dfrac{t \cdot f}{N_{\mathrm{div}}}=\dfrac{1.748\,s \cdot 2.4 \cdot 10^{9}\,\mathrm{cy}}{2 \cdot 10^{9}\,s}\approx 2\,\dfrac{\mathrm{cy}}{\mathrm{div}}">
* `-O1 -no-vec`  
   Runtime `t=3.407s`  
   Solution: <img src="https://render.githubusercontent.com/render/math?math=\dfrac{\mathrm{cy}}{N_{\mathrm{div}}}=\dfrac{t \cdot f}{N_{\mathrm{div}}}=\dfrac{3.407\,s \cdot 2.4 \cdot 10^{9}\,\mathrm{cy}}{2 \cdot 10^{9}\,s}\approx 4\,\dfrac{\mathrm{cy}}{\mathrm{div}}">

### Did you expect this result?

Not fully pipelined.  Long latency instruction.

SIMD vectorisation: Factor of two from scalar to SSE, No (significant) further improvement going from 128bit to 256bit to 512bit SIMD.


### Advanced Assembly analysis
OSACA Output for AVX-512 (4x unrolling and 8x SIMD vectorized):
```
osaca --arch icl divAVX.s
Open Source Architecture Code Analyzer (OSACA) - 0.4.7
Analyzed file:      divAVX.s
Architecture:       ICL
Timestamp:          2022-03-25 13:10:15


 P - Throughput of LOAD operation can be hidden behind a past or future STORE instruction
 * - Instruction micro-ops not bound to a port
 X - No throughput/latency information for this instruction in data file


Combined Analysis Report
------------------------
                                                Port pressure in cycles
     |  0   -  0DV  |  1   - 1DV  |  2   -  2D  |  3   -  3D  |  4   |  5   |  6   |  7   |  8   |  9   ||  CP  | LCD  |
------------------------------------------------------------------------------------------------------------------------
   1 |              |             |             |             |      |      |      |      |      |      ||      |      |   .LBB1_1:                                # =>This Inner Loop Header: Depth=1
   2 |              |             |             |             |      | 1.00 |      |      |      |      ||  5.0 |      |   vpbroadcastd    %eax, %zmm8
   3 | 0.00         |             |             |             |      | 1.00 |      |      |      |      ||  1.0 |      |   vpord   %zmm1, %zmm8, %zmm8
   4 | 1.00         |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vcvtdq2pd   %ymm8, %zmm9
   5 |              |             |             |             |      | 1.00 |      |      |      |      ||  3.0 |      |   vextracti64x4   $1, %zmm8, %ymm8
   6 | 1.00         |             |             |             |      | 1.00 |      |      |      |      ||  7.0 |      |   vcvtdq2pd   %ymm8, %zmm8
   7 | 1.00         |             |             |             |      |      |      |      |      |      ||  4.0 |      |   vfmadd213pd %zmm3, %zmm2, %zmm8     # zmm8 = (zmm2 * zmm8) + zmm3
   8 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vfmadd213pd %zmm3, %zmm2, %zmm9     # zmm9 = (zmm2 * zmm9) + zmm3
   9 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vfmadd213pd %zmm4, %zmm9, %zmm9     # zmm9 = (zmm9 * zmm9) + zmm4
  10 | 1.00         |             |             |             |      |      |      |      |      |      ||  4.0 |      |   vfmadd213pd %zmm4, %zmm8, %zmm8     # zmm8 = (zmm8 * zmm8) + zmm4
  11 |              | 1.00        |             |             |      | 0.00 |      |      |      |      ||      |      |   leal    16(%rax), %ecx
  12 |              |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vpbroadcastd    %ecx, %zmm10
  13 | 0.00         |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vpord   %zmm1, %zmm10, %zmm10
  14 | 2.00   16.00 |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vdivpd  %zmm9, %zmm5, %zmm9
  15 |              |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vextracti64x4   $1, %zmm10, %ymm11
  16 | 1.00         |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vcvtdq2pd   %ymm11, %zmm11
  17 | 2.00   16.00 |             |             |             |      | 1.00 |      |      |      |      || 23.0 |      |   vdivpd  %zmm8, %zmm5, %zmm8
  18 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vaddpd  %zmm6, %zmm9, %zmm6
  19 | 1.00         |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vcvtdq2pd   %ymm10, %zmm9
  20 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vfmadd213pd %zmm3, %zmm2, %zmm9     # zmm9 = (zmm2 * zmm9) + zmm3
  21 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vfmadd213pd %zmm3, %zmm2, %zmm11    # zmm11 = (zmm2 * zmm11) + zmm3
  22 | 1.00         |             |             |             |      |      |      |      |      |      ||  4.0 |  4.0 |   vaddpd  %zmm7, %zmm8, %zmm7
  23 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vfmadd213pd %zmm4, %zmm11, %zmm11   # zmm11 = (zmm11 * zmm11) + zmm4
  24 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vfmadd213pd %zmm4, %zmm9, %zmm9     # zmm9 = (zmm9 * zmm9) + zmm4
  25 | 2.00   16.00 |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vdivpd  %zmm9, %zmm5, %zmm8
  26 | 1.00         |             |             |             |      |      |      |      |      |      ||      |      |   vaddpd  %zmm6, %zmm8, %zmm6
  27 | 2.00   16.00 |             |             |             |      | 1.00 |      |      |      |      ||      |      |   vdivpd  %zmm11, %zmm5, %zmm8
  28 | 1.00         |             |             |             |      |      |      |      |      |      ||  4.0 |  4.0 |   vaddpd  %zmm7, %zmm8, %zmm7
  29 | 0.00         | 0.50        |             |             |      | 0.00 | 0.50 |      |      |      ||      |      |   addl    $32, %eax
  30 | 0.00         | 0.25        |             |             |      | 0.00 | 0.75 |      |      |      ||      |      |   cmpl    $1999999984, %ecx               # imm = 0x773593F0
  31 |              |             |             |             |      |      |      |      |      |      ||      |      | * jb  .LBB1_1

       24.0   64.00   1.75                                             14.0   1.25                           55    8.0




Loop-Carried Dependencies Analysis Report
-----------------------------------------
  22 |  8.0 | vaddpd  %zmm7, %zmm8, %zmm7         | [22, 28]
  18 |  8.0 | vaddpd  %zmm6, %zmm9, %zmm6         | [18, 26]
  29 |  1.0 | addl    $32, %eax                   | [29]
```
